#include "render.h"
#include "arpa/inet.h"
#include "pipeline.h"
#include "vk_mem_alloc.h"
#include "vulkan/vk_enum_string_helper.h"
#include "vulkan/vulkan_core.h"

#include "ft2build.h"
#include FT_FREETYPE_H

ColoredRect colored_rect(float width, float height, float r, float g, float b, float a, float x, float y, float z) {
  ColoredRect rect = {
    .size = {width, height},
    .pos = {x, y, z},
    .color = {r, g, b, a},
  };
  return rect;
}

VkResult render_thread(GLFWwindow* window, RenderContext* render_context) {
  VkResult result;
  UIContext ui_context;
  VkBuffer colored_rect_buffer;
  VmaAllocation colored_rect_memory;

  result = init_pipelines(render_context->device, render_context->allocator, render_context->swapchain_extent, render_context->window_scale, render_context->render_pass, render_context->transfer_queue, render_context->transfer_pool, &ui_context);
  if(result != VK_SUCCESS) {
    return result;
  }

  FT_Library library;
  if(FT_Init_FreeType(&library) != FT_Err_Ok) {
    return VK_ERROR_UNKNOWN;
  }

  FontDescriptor test_font;
  uint32_t* charmap;
  result = load_font(render_context->device, render_context->allocator, ui_context.font_layout, ui_context.font_pool, render_context->transfer_pool, render_context->transfer_queue, library, "test.ttf", 16, VK_TRUE, &charmap, &test_font);
  if(result != VK_SUCCESS) {
    return result;
  }

  VkBufferCreateInfo colored_rect_buffer_info = {
    .sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO,
    .usage = VK_BUFFER_USAGE_VERTEX_BUFFER_BIT,
    .size = 3*sizeof(ColoredRect),
    .sharingMode = VK_SHARING_MODE_EXCLUSIVE,
  };

  VmaAllocationCreateInfo colored_rect_memory_info = {
    .usage = VMA_MEMORY_USAGE_CPU_TO_GPU,
  };

  result = vmaCreateBuffer(render_context->allocator, &colored_rect_buffer_info, &colored_rect_memory_info, &colored_rect_buffer, &colored_rect_memory, NULL);
  if(result != VK_SUCCESS) {
    return result;
  }

  ColoredRect* colored_rects;

  result = vmaMapMemory(render_context->allocator, colored_rect_memory, (void**)&colored_rects);
  if(result != VK_SUCCESS) {
    return result;
  }

  colored_rects[0] = colored_rect(100.0, 100.0, 1.0, 0.0, 0.0, 1.0,   0.0,   0.0, 0.5);
  colored_rects[1] = colored_rect(100.0, 100.0, 0.0, 1.0, 0.0, 1.0,   0.0, 100.0, 0.5);
  colored_rects[2] = colored_rect(100.0, 100.0, 0.0, 0.0, 1.0, 1.0, 100.0,   0.0, 0.5);

  vmaUnmapMemory(render_context->allocator, colored_rect_memory);

  TextPointersMemory text_pointers;
  VkDeviceAddress text_pointers_address;
  result = create_text_pointers(2, 100, render_context->device, render_context->allocator, render_context->transfer_pool, render_context->transfer_queue, &text_pointers, &text_pointers_address);
  if(result != VK_SUCCESS) {
    return result;
  }

  VkBuffer temp_buffer;
  VmaAllocation temp_memory;
  void* mapped;
  result = create_transfer_buffer(2*sizeof(String) + 100*sizeof(uint32_t), render_context->allocator, &temp_buffer, &temp_memory, (void**)&mapped);
  if(result != VK_SUCCESS) {
    return result;
  }

  String* mapped_strings = (String*)mapped;
  mapped_strings[0].pos[0] = 0;
  mapped_strings[0].pos[1] = 200.0;
  mapped_strings[0].pos[2] = 0.2;
  mapped_strings[0].color[0] = 1.0;
  mapped_strings[0].color[1] = 1.0;
  mapped_strings[0].color[2] = 1.0;
  mapped_strings[0].color[3] = 1.0;
  mapped_strings[0].size = 50.0;
  mapped_strings[0].length = 50;
  mapped_strings[0].offset = 0;

  mapped_strings[1].pos[0] = 0;
  mapped_strings[1].pos[1] = 250.0;
  mapped_strings[1].pos[2] = 0.1;
  mapped_strings[1].color[0] = 1.0;
  mapped_strings[1].color[1] = 1.0;
  mapped_strings[1].color[2] = 1.0;
  mapped_strings[1].color[3] = 1.0;
  mapped_strings[1].size = 50.0;
  mapped_strings[1].length = 50;
  mapped_strings[1].offset = 50;

  uint32* mapped_codes = (uint32_t*)(mapped + 2*sizeof(String));
  for(uint32_t i = 0; i < 100; i++) {
    mapped_codes[i]  = i;
  }

  VkCommandBuffer command_buffer = command_begin_single(render_context->device, render_context->transfer_pool);
  VkBufferCopy copy_string = {
    .size = 2*sizeof(String),
    .srcOffset = 0,
  };
  vkCmdCopyBuffer(command_buffer, temp_buffer, text_pointers.strings_buffer, 1, &copy_string);
  VkBufferCopy copy_codes = {
    .size = 100*sizeof(uint32_t),
    .srcOffset = 2*sizeof(String),
  };
  vkCmdCopyBuffer(command_buffer, temp_buffer, text_pointers.codes_buffer, 1, &copy_codes);
  result = command_end_single(render_context->device, command_buffer, render_context->transfer_pool, render_context->transfer_queue);
  if(result != VK_SUCCESS) {
    return result;
  }

  vkQueueWaitIdle(render_context->transfer_queue.handle);
  vmaUnmapMemory(render_context->allocator, temp_memory);
  vmaDestroyBuffer(render_context->allocator, temp_buffer, temp_memory);

  UILayer test_layer = {
    .colored_rects = colored_rect_buffer,
    .colored_rect_count = 3,

    .font = test_font,
    .chars = text_pointers.characters_buffer,
    .string_count = 2,
    .chars_count = 100,
    .string_pointers = text_pointers_address,
    .string_draw = text_pointers.draw_buffer,
  };

  while(glfwWindowShouldClose(window) == 0) {
    glfwPollEvents();

    result = draw_frame(render_context, &ui_context, &test_layer, 1);
    if(result != VK_SUCCESS) {
      fprintf(stderr, "draw_frame error: %s\n", string_VkResult(result));
      return result;
    }
  }
  return 0;
}

int logic_thread() {
  return 0;
}

int network_thread() {
  return 0;
}

int main() {
  GLFWwindow* window = init_window();
  if(window == NULL) {
    return 1;
  }

  RenderContext render_context = {};
  if(init_vulkan(window, &render_context) != VK_SUCCESS) {
    return 2;
  }

  if(render_thread(window, &render_context) != VK_SUCCESS) {
    return 3;
  }

  return 0;
}
